# coding=utf-8
# Copyright 2023 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""optimum.exporters command-line interface base classes."""

from .. import BaseOptimumCLICommand, CommandInfo
from .onnx import ONNXExportCommand
from .tflite import TFLiteExportCommand


class ExportCommand(BaseOptimumCLICommand):
    COMMAND = CommandInfo(
        name="export",
        help="Export PyTorch and TensorFlow models to several format.",
    )
    SUBCOMMANDS = (
        CommandInfo(
            name="onnx",
            help="Export PyTorch and TensorFlow to ONNX.",
            subcommand_class=ONNXExportCommand,
        ),
        CommandInfo(
            name="tflite",
            help="Export TensorFlow to TensorFlow Lite.",
            subcommand_class=TFLiteExportCommand,
        ),
    )
